'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

exports.default = function (_ref) {
    var t = _ref.types;

    var IGNORE_REASSIGNMENT_SYMBOL = Symbol();

    var reassignmentVisitor = {
        'AssignmentExpression|UpdateExpression': function AssignmentExpressionUpdateExpression(path) {
            if (path.node[IGNORE_REASSIGNMENT_SYMBOL]) return;
            path.node[IGNORE_REASSIGNMENT_SYMBOL] = true;

            var arg = path.get(path.isAssignmentExpression() ? 'left' : 'argument');
            if (!arg.isIdentifier()) return;

            var name = arg.node.name;

            // redeclared in this scope
            if (this.scope.getBinding(name) !== path.scope.getBinding(name)) return;

            var exportedNames = this.exports[name];
            if (!exportedNames) return;

            var node = path.node;

            // if it is a non-prefix update expression (x++ etc)
            // then we must replace with the expression (_export('x', x + 1), x++)
            // in order to ensure the same update expression value
            var isPostUpdateExpression = path.isUpdateExpression() && !node.prefix;
            if (isPostUpdateExpression) {
                if (node.operator === '++') {
                    node = t.binaryExpression('+', node.argument, t.numericLiteral(1));
                } else if (node.operator === '--') {
                    node = t.binaryExpression('-', node.argument, t.numericLiteral(1));
                } else {
                    isPostUpdateExpression = false;
                }
            }

            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
                for (var _iterator = exportedNames[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var exportedName = _step.value;

                    node = this.buildCall(exportedName, node).expression;
                }
            } catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                } finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }

            if (isPostUpdateExpression) {
                node = t.sequenceExpression([node, path.node]);
            }

            path.replaceWith(node);
        }
    };

    return {
        visitor: {
            ReferencedIdentifier: function ReferencedIdentifier(path, state) {
                if (path.node.name == '__moduleName' && !path.scope.hasBinding('__moduleName')) {
                    path.replaceWith(t.memberExpression(state.contextIdent, t.identifier('id')));
                }
            },


            Program: {
                enter: function enter(path, state) {
                    state.contextIdent = path.scope.generateUidIdentifier('context');
                },
                exit: function exit(path, state) {
                    var exportIdent = path.scope.generateUidIdentifier('export');
                    var contextIdent = state.contextIdent;

                    var exportNames = Object.create(null);
                    var modules = Object.create(null);

                    var beforeBody = [];
                    var setters = [];
                    var sources = [];
                    var variableIds = [];
                    var removedPaths = [];

                    function addExportName(key, val) {
                        exportNames[key] = exportNames[key] || [];
                        exportNames[key].push(val);
                    }

                    function pushModule(source, key, specifiers) {
                        var _modules = modules[source] = modules[source] || { imports: [], exports: [] };
                        _modules[key] = _modules[key].concat(specifiers);
                    }

                    function buildExportCall(name, val) {
                        return t.expressionStatement(t.callExpression(exportIdent, [t.stringLiteral(name), val]));
                    }

                    var body = path.get('body');

                    var canHoist = true;
                    var _iteratorNormalCompletion2 = true;
                    var _didIteratorError2 = false;
                    var _iteratorError2 = undefined;

                    try {
                        for (var _iterator2 = body[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                            var _path = _step2.value;

                            if (_path.isExportDeclaration()) _path = _path.get('declaration');
                            if (_path.isVariableDeclaration() && _path.node.kind !== 'var') {
                                canHoist = false;
                                break;
                            }
                        }
                    } catch (err) {
                        _didIteratorError2 = true;
                        _iteratorError2 = err;
                    } finally {
                        try {
                            if (!_iteratorNormalCompletion2 && _iterator2.return) {
                                _iterator2.return();
                            }
                        } finally {
                            if (_didIteratorError2) {
                                throw _iteratorError2;
                            }
                        }
                    }

                    var _iteratorNormalCompletion3 = true;
                    var _didIteratorError3 = false;
                    var _iteratorError3 = undefined;

                    try {
                        for (var _iterator3 = body[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                            var _path2 = _step3.value;

                            if (canHoist && _path2.isFunctionDeclaration()) {
                                beforeBody.push(_path2.node);
                                removedPaths.push(_path2);
                            } else if (_path2.isImportDeclaration()) {
                                var _source = _path2.node.source.value;
                                pushModule(_source, 'imports', _path2.node.specifiers);
                                for (var name in _path2.getBindingIdentifiers()) {
                                    _path2.scope.removeBinding(name);
                                    variableIds.push(t.identifier(name));
                                }
                                _path2.remove();
                            } else if (_path2.isExportAllDeclaration()) {
                                pushModule(_path2.node.source.value, 'exports', _path2.node);
                                _path2.remove();
                            } else if (_path2.isExportDefaultDeclaration()) {
                                var declar = _path2.get('declaration');
                                if (declar.isClassDeclaration() || declar.isFunctionDeclaration()) {
                                    var id = declar.node.id;
                                    var nodes = [];

                                    if (id) {
                                        nodes.push(declar.node);
                                        nodes.push(buildExportCall('default', id));
                                        addExportName(id.name, 'default');
                                    } else {
                                        nodes.push(buildExportCall('default', t.toExpression(declar.node)));
                                    }

                                    if (!canHoist || declar.isClassDeclaration()) {
                                        _path2.replaceWithMultiple(nodes);
                                    } else {
                                        beforeBody = beforeBody.concat(nodes);
                                        removedPaths.push(_path2);
                                    }
                                } else {
                                    _path2.replaceWith(buildExportCall('default', declar.node));
                                }
                            } else if (_path2.isExportNamedDeclaration()) {
                                var _declar = _path2.get('declaration');

                                if (_declar.node) {
                                    _path2.replaceWith(_declar);

                                    var _nodes = [];
                                    var bindingIdentifiers = void 0;
                                    if (_path2.isFunction()) {
                                        var _node = _declar.node;
                                        var _name = _node.id.name;
                                        if (canHoist) {
                                            addExportName(_name, _name);
                                            beforeBody.push(_node);
                                            beforeBody.push(buildExportCall(_name, _node.id));
                                            removedPaths.push(_path2);
                                        } else {
                                            bindingIdentifiers = _defineProperty({}, _name, _node.id);
                                        }
                                    } else {
                                        bindingIdentifiers = _declar.getBindingIdentifiers();
                                    }
                                    for (var _name2 in bindingIdentifiers) {
                                        addExportName(_name2, _name2);
                                        _nodes.push(buildExportCall(_name2, t.identifier(_name2)));
                                    }
                                    _path2.insertAfter(_nodes);
                                } else {
                                    var _specifiers = _path2.node.specifiers;
                                    if (_specifiers && _specifiers.length) {
                                        if (_path2.node.source) {
                                            pushModule(_path2.node.source.value, 'exports', _specifiers);
                                            _path2.remove();
                                        } else {
                                            var _nodes2 = [];

                                            var _iteratorNormalCompletion7 = true;
                                            var _didIteratorError7 = false;
                                            var _iteratorError7 = undefined;

                                            try {
                                                for (var _iterator7 = _specifiers[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
                                                    var _specifier = _step7.value;

                                                    _nodes2.push(buildExportCall(_specifier.exported.name, _specifier.local));
                                                    addExportName(_specifier.local.name, _specifier.exported.name);
                                                }
                                            } catch (err) {
                                                _didIteratorError7 = true;
                                                _iteratorError7 = err;
                                            } finally {
                                                try {
                                                    if (!_iteratorNormalCompletion7 && _iterator7.return) {
                                                        _iterator7.return();
                                                    }
                                                } finally {
                                                    if (_didIteratorError7) {
                                                        throw _iteratorError7;
                                                    }
                                                }
                                            }

                                            _path2.replaceWithMultiple(_nodes2);
                                        }
                                    }
                                }
                            }
                        }
                    } catch (err) {
                        _didIteratorError3 = true;
                        _iteratorError3 = err;
                    } finally {
                        try {
                            if (!_iteratorNormalCompletion3 && _iterator3.return) {
                                _iterator3.return();
                            }
                        } finally {
                            if (_didIteratorError3) {
                                throw _iteratorError3;
                            }
                        }
                    }

                    for (var source in modules) {
                        var specifiers = modules[source];

                        var setterBody = [];
                        var target = path.scope.generateUidIdentifier(source);

                        var _iteratorNormalCompletion4 = true;
                        var _didIteratorError4 = false;
                        var _iteratorError4 = undefined;

                        try {
                            for (var _iterator4 = specifiers.imports[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
                                var specifier = _step4.value;

                                if (t.isImportNamespaceSpecifier(specifier)) {
                                    setterBody.push(t.expressionStatement(t.assignmentExpression('=', specifier.local, target)));
                                } else if (t.isImportDefaultSpecifier(specifier)) {
                                    specifier = t.importSpecifier(specifier.local, t.identifier('default'));
                                }

                                if (t.isImportSpecifier(specifier)) {
                                    setterBody.push(t.expressionStatement(t.assignmentExpression('=', specifier.local, t.memberExpression(target, specifier.imported))));
                                }
                            }
                        } catch (err) {
                            _didIteratorError4 = true;
                            _iteratorError4 = err;
                        } finally {
                            try {
                                if (!_iteratorNormalCompletion4 && _iterator4.return) {
                                    _iterator4.return();
                                }
                            } finally {
                                if (_didIteratorError4) {
                                    throw _iteratorError4;
                                }
                            }
                        }

                        if (specifiers.exports.length) {
                            var exportObjRef = path.scope.generateUidIdentifier('exportObj');

                            setterBody.push(t.variableDeclaration('var', [t.variableDeclarator(exportObjRef, t.objectExpression([]))]));

                            var _iteratorNormalCompletion5 = true;
                            var _didIteratorError5 = false;
                            var _iteratorError5 = undefined;

                            try {
                                for (var _iterator5 = specifiers.exports[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
                                    var node = _step5.value;

                                    if (t.isExportAllDeclaration(node)) {
                                        setterBody.push(buildExportAll({
                                            KEY: path.scope.generateUidIdentifier('key'),
                                            EXPORT_OBJ: exportObjRef,
                                            TARGET: target
                                        }));
                                    } else if (t.isExportSpecifier(node)) {
                                        setterBody.push(t.expressionStatement(t.assignmentExpression('=', t.memberExpression(exportObjRef, node.exported), t.memberExpression(target, node.local))));
                                    } else {
                                        // todo
                                    }
                                }
                            } catch (err) {
                                _didIteratorError5 = true;
                                _iteratorError5 = err;
                            } finally {
                                try {
                                    if (!_iteratorNormalCompletion5 && _iterator5.return) {
                                        _iterator5.return();
                                    }
                                } finally {
                                    if (_didIteratorError5) {
                                        throw _iteratorError5;
                                    }
                                }
                            }

                            setterBody.push(t.expressionStatement(t.callExpression(exportIdent, [exportObjRef])));
                        }

                        sources.push(t.stringLiteral(source));
                        setters.push(t.stringLiteral(setterBody[0].expression.left.name));
                    }

                    var moduleName = this.getModuleName();
                    if (moduleName) moduleName = t.stringLiteral(moduleName);

                    if (canHoist) {
                        (0, _babelHelperHoistVariables2.default)(path, function (id) {
                            return variableIds.push(id);
                        });
                    }

                    if (variableIds.length) {
                        beforeBody.unshift(t.variableDeclaration('var', variableIds.map(function (id) {
                            return t.variableDeclarator(id);
                        })));
                    }

                    path.traverse(reassignmentVisitor, {
                        exports: exportNames,
                        buildCall: buildExportCall,
                        scope: path.scope
                    });

                    var _iteratorNormalCompletion6 = true;
                    var _didIteratorError6 = false;
                    var _iteratorError6 = undefined;

                    try {
                        for (var _iterator6 = removedPaths[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
                            var _path3 = _step6.value;

                            _path3.remove();
                        }
                    } catch (err) {
                        _didIteratorError6 = true;
                        _iteratorError6 = err;
                    } finally {
                        try {
                            if (!_iteratorNormalCompletion6 && _iterator6.return) {
                                _iterator6.return();
                            }
                        } finally {
                            if (_didIteratorError6) {
                                throw _iteratorError6;
                            }
                        }
                    }

                    var variables = {
                        type: 'Identifier',
                        name: ''
                    };

                    moduleName = t.stringLiteral(_path5.default.basename(this.file.opts.filename).replace('.js', ''));

                    setters.forEach(function (item, index) {

                        variables.name += item.value;

                        if (index !== setters.length - 1) {
                            variables.name += ', ';
                        }
                    });

                    path.node.body = [buildTemplate({
                        SYSTEM_REGISTER: t.memberExpression(t.identifier(state.opts.systemGlobal || 'Patchwork'), t.identifier('weave')),
                        BEFORE_BODY: beforeBody,
                        MODULE_NAME: moduleName,
                        SOURCES: sources,
                        BODY: path.node.body,
                        VARIABLES: variables
                    })];
                }
            }
        }
    };
};

var _babelHelperHoistVariables = require('babel-helper-hoist-variables');

var _babelHelperHoistVariables2 = _interopRequireDefault(_babelHelperHoistVariables);

var _babelTemplate = require('babel-template');

var _babelTemplate2 = _interopRequireDefault(_babelTemplate);

var _path4 = require('path');

var _path5 = _interopRequireDefault(_path4);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; } /* eslint max-len: 0 */

var buildTemplate = (0, _babelTemplate2.default)('\n  SYSTEM_REGISTER(MODULE_NAME, [SOURCES], function (VARIABLES) {\n    "use strict";\n    BEFORE_BODY;\n    BODY;\n  });\n');

var buildExportAll = (0, _babelTemplate2.default)('\n  for (var KEY in TARGET) {\n    if (KEY !== "default" && KEY !== "__esModule") EXPORT_OBJ[KEY] = TARGET[KEY];\n  }\n');